#include "testlib.h"
#include "gen_lib.h"
#include <bits/stdc++.h>

using namespace std;

constexpr int N = 3'000'00;
constexpr int SMALL_N = 3000;
constexpr int GROUPS = 10;
constexpr int MAX_N_M[GROUPS + 1]{/* 0:  */ N,
                                  /* 1:  */ SMALL_N,
                                  /* 2:  */ SMALL_N,
                                  /* 3:  */ SMALL_N,
                                  /* 4:  */ SMALL_N,
                                  /* 5:  */ SMALL_N,
                                  /* 6:  */ SMALL_N,
                                  /* 7:  */ SMALL_N,
                                  /* 8:  */ SMALL_N,
                                  /* 9:  */ N,
                                  /* 10: */ N};

ArgumentParser parser;
Logger logger;

struct all_test_cases {
    struct test_case {
        int n;
        vector<pair<int, int>> edges1;
        vector<pair<int, int>> edges2;

        void create_same_edge() {
            auto check = [](const vector<pair<int, int>> &edges) {
                if (find(edges.begin(), edges.end(), pair<int, int>{0, 1}) != edges.end()) {
                    return true;
                }
                return find(edges.begin(), edges.end(), pair<int, int>{1, 0}) != edges.end();
            };

            if (check(edges1) && check(edges2)) {
                return;
            }

            for (auto &[v, u] : edges1) {
                if (v > u) {
                    swap(v, u);
                }
            }
            sort(edges1.begin(), edges1.end());
            random::shuffle(edges2.begin(), edges2.end(), rnd);

            pair<int, int> same{-1, -1};
            for (auto [v, u] : edges2) {
                if (binary_search(edges1.begin(), edges1.end(), pair<int, int>{min(v, u), max(v, u)})) {
                    same = {v, u};
                }
            }

            if (same.first == -1) {
                edges1.push_back(edges2[0]);
                same = edges2[0];
            }

            vector<int> perm(n);
            iota(perm.begin(), perm.end(), 0);
            swap(perm[same.first], perm[0]);
            if (perm[1] == 1) {
                swap(perm[same.second], perm[1]);
            } else {
                assert(perm[0] == 1);
                swap(perm[same.second], perm[0]);
            }

            assert(perm[same.first] == 0);
            assert(perm[same.second] == 1);

            for (auto &[v, u] : edges1) {
                v = perm[v];
                u = perm[u];
            }
            for (auto &[v, u] : edges2) {
                v = perm[v];
                u = perm[u];
            }
        }

        void print(int group) {
            if (group == 7) {
                create_same_edge();
            }

            shuffle(edges1.begin(), edges1.end());
            for (auto &[v, u] : edges1) {
                if (rnd.next(0, 1)) swap(v, u);
            }

            if (group != 9 && group != 5) {
                shuffle(edges2.begin(), edges2.end());
                for (auto &[v, u] : edges2) {
                    if (rnd.next(0, 1)) swap(v, u);
                }
            } else {
                for (auto &[v, u] : edges2) {
                    if (v > u) swap(v, u);
                }
                sort(edges2.begin(), edges2.end());
            }

            println(n, edges1.size(), edges2.size());
            for (auto edges : {edges1, edges2}) {
                for (auto [v, u] : edges) {
                    println(v + 1, u + 1);
                }
            }
        }
    };

    int group{-1};
    int MAX{-1};

    vector<test_case> test_cases;
    int sum_n{0};
    int sum_m1{0};
    int sum_m2{0};

    bool add_test_case(int n, const vector<pair<int, int>> &edges1, const vector<pair<int, int>> &edges2) {
        if (sum_n + n > MAX || sum_m1 + int(edges1.size()) > MAX || sum_m2 + int(edges2.size()) > MAX) {
            return false;
        }
        sum_n += n;
        sum_m1 += edges1.size();
        sum_m2 += edges2.size();
        test_cases.push_back({n, edges1, edges2});
        return true;
    }

    void print() {
        cout << test_cases.size() << ' ' << group << '\n';
        for (auto &test_case : test_cases) {
            test_case.print(group);
        }
    }
};

all_test_cases result;

bool validate_first_graph(const UndirectedGraph &graph1) {
    if (result.group == 3 && graph1.components() != 2) {
        return false;
    }

    if (result.group == 4) {
        for (auto x : graph1.getDegreeSequence()) {
            if (x < 1) {
                return false;
            }
        }
    }

    if (result.group == 7 && !graph1.containsEdge({0, 1})) {
        return false;
    }
    return true;
}

bool validate_second_graph(const UndirectedGraph &graph2) {
    if (graph2.components() != 1) {
        return false;
    }

    if (result.group == 2 && graph2.size().second != 1ll * graph2.size().first * (graph2.size().first - 1) / 2) {
        return false;
    }

    if (result.group == 9) {
        if (!graph2.isTree()) {
            return false;
        }
        for (int i = 0; i + 1 < int(graph2.size().first); i++) {
            if (!graph2.containsEdge({i, i + 1})) {
                return false;
            }
        }
    }

    if (result.group == 5) {
        if (!graph2.isTree()) {
            return false;
        }
        for (int i = 1; i < int(graph2.size().first); i++) {
            if (!graph2.containsEdge({0, i})) {
                return false;
            }
        }
    }

    if (result.group == 6 && !graph2.isTree()) {
        return false;
    }

    if (result.group == 7 && !graph2.containsEdge({0, 1})) {
        return false;
    }
    return true;
}

vector<pair<int, int>> check_first_graph_edges(int n, vector<pair<int, int>> edges) {
    if (result.group != 4) {
        return edges;
    }
    vector<int> degs(n);
    for (auto [v, u] : edges) {
        degs[v]++;
        degs[u]++;
    }
    vector<int> isolated;
    for (int i = 0; i < n; i++) {
        if (degs[i] == 0) {
            isolated.push_back(i);
        }
    }
    if (isolated.empty()) {
        return edges;
    }
    if (isolated.size() == 1) {
        int v;
        do {
            v = random::random(0, n, rnd);
        } while (v == isolated[0]);
        edges.emplace_back(isolated[0], v);
        return edges;
    }

    random::shuffle(isolated.begin(), isolated.end(), rnd);
    for (int i = 0; i < int(isolated.size()); i++) {
        if (i % 2 == 1 || i == int(isolated.size()) - 1) {
            edges.emplace_back(isolated[i - 1], isolated[i]);
        }
    }
    return edges;
}

vector<pair<int, int>> generate_first_graph(int n) {
    const auto first_mode = parser.getString("first_mode");
    if (first_mode == "forest") {
        const int max_m = min<int>(n - 1, n * parser.getDouble("edges_frac"));
        int m = random::random(0, max_m + 1, rnd);
        return check_first_graph_edges(n, treesGenerators::randomPruferForest(n, m, rnd).getEdges());
    } else if (first_mode == "random") {
        const int max_m = n * parser.getDouble("edges_frac");
        int m = random::random(0, max_m + 1, rnd);
        return check_first_graph_edges(n, undirectedGraphsGenerators::randomNoLoopsAndMultipleEdges(n, m, rnd).getEdges());
    } else if (first_mode == "clique") {
        vector<pair<int, int>> edges;
        edges.reserve((n - 1) * (n - 2) / 2);
        int center = random::random(0, n, rnd);
        for (int i = 0; i < n; i++) {
            for (int j = i + 1; j < n; j++) {
                if (i != center && j != center) {
                    edges.emplace_back(i, j);
                }
            }
        }
        return check_first_graph_edges(n, edges);
    } else if (first_mode == "many_cliques") {
        const int cliques = parser.getInt("cliques");
        auto distrib = random::distributeToGroups(n, 0, cliques, rnd);

        vector<pair<int, int>> edges;
        edges.reserve(n * (n - 1) / 2);
        for (int i = 0; i < n; i++) {
            for (int j = i + 1; j < n; j++) {
                if (distrib[i] == distrib[j]) {
                    edges.emplace_back(i, j);
                }
            }
        }
        return check_first_graph_edges(n, edges);
    } else if (first_mode == "fixed_forest") {
        int comps;
        if (parser.hasParameter("comps")) {
            comps = parser.getInt("comps");
        } else {
            comps = max(2, min<int>(n, n * parser.getDouble("comps_frac")));
        }
        auto distrib = random::distributeToGroups(n, 0, comps, rnd);
        vector<vector<int>> groups(comps);
        for (int i = 0; i < n; i++) {
            groups[distrib[i]].push_back(i);
        }

        vector<pair<int, int>> edges;
        for (const auto &group : groups) {
            for (const auto &[v, u] : treesGenerators::randomPruferCode(group.size(), rnd).getEdges()) {
                edges.emplace_back(group[v], group[u]);
            }
        }
        return check_first_graph_edges(n, edges);
    } else if (first_mode == "random_comps") {
        const int comps = parser.getInt("comps");
        const double extra_edges_frac = parser.getDouble("extra_edges_frac");

        auto distrib = random::distributeToGroups(n, 0, comps, rnd);
        vector<vector<int>> groups(comps);
        for (int i = 0; i < n; i++) {
            groups[distrib[i]].push_back(i);
        }

        vector<pair<int, int>> edges;
        for (const auto &group : groups) {
            auto cur_edges = treesGenerators::randomPruferCode(group.size(), rnd).getEdges();
            for (auto edge : cur_edges) {
                edges.emplace_back(group[edge.first], group[edge.second]);
                assert(group[edge.first] != group[edge.second]);
            }

            set<pair<int, int>> was;
            for (const auto &[v, u] : cur_edges) {
                was.insert(minmax(v, u));
            }

            for (int it = 0; it < int(group.size() * extra_edges_frac); it++) {
                int v, u;
                do {
                    v = random::random<int>(0, group.size(), rnd);
                    u = random::random<int>(0, group.size(), rnd);
                } while (v == u);
                if (was.insert(minmax(v, u)).second) {
                    edges.emplace_back(group[v], group[u]);
                    assert(group[v] != group[u]);
                }
            }
        }
        return check_first_graph_edges(n, edges);
    } else if (first_mode == "pairing") {
        vector<pair<int, int>> edges;
        for (int i = 0; i + 1 < n; i += 2) {
            edges.emplace_back(i, i + 1);
        }
        if (n % 2 == 1) {
            edges.emplace_back(n - 2, n - 1);
        }
        return check_first_graph_edges(n, edges);
    }
    global::logAndThrowException("Incorrect generator mode for the first graph: " + first_mode);
    return {};
}

vector<pair<int, int>> generate_second_graph(int n) {
    const auto second_mode = parser.getString("second_mode");
    if (second_mode == "bamboo") {
        vector<pair<int, int>> edges(n - 1);
        for (int i = 0; i + 1 < n; i++) {
            edges[i] = {i, i + 1};
        }
        return edges;
    } else if (second_mode == "star") {
        vector<pair<int, int>> edges(n - 1);
        for (int i = 1; i < n; i++) {
            edges[i - 1] = {0, i};
        }
        return edges;
    } else if (second_mode == "tree") {
        return treesGenerators::randomPruferCode(n, rnd).getEdges();
    } else if (second_mode == "complete") {
        vector<pair<int, int>> edges;
        edges.reserve(n * (n - 1) / 2);
        for (int i = 0; i < n; i++) {
            for (int j = i + 1; j < n; j++) {
                edges.emplace_back(i, j);
            }
        }
        return edges;
    } else if (second_mode == "extra_tree") {
        const int extra = n * parser.getDouble("extra_frac");
        auto edges = treesGenerators::randomPruferCode(n, rnd).getEdges();
        set<pair<int, int>> was;
        for (auto [v, u] : edges) {
            was.insert(minmax(v, u));
        }
        for (int i = 0; i < extra; i++) {
            int v, u;
            do {
                v = random::random(0, n, rnd);
                u = random::random(0, n, rnd);
            } while (v == u);
            if (was.insert(minmax(v, u)).second) {
                edges.emplace_back(v, u);
            }
        }
        return edges;
    }
    global::logAndThrowException("Incorrect generator mode for the first graph: " + second_mode);
    return {};
}

void gen_all_possible_cases() {
    int n = parser.getInt("n");
    int start = parser.getInt("graph_start", 0);

    int index = 0;
    int connected_index = 0;
    undirectedGraphsGenerators::AllUndirectedGraphsGenerator generator1(n);
    bool fail = false;

    do {
        auto first_graph = generator1.buildGraph();
        if (!validate_first_graph(first_graph)) {
            continue;
        }
        int first_components = first_graph.components();
        auto edges1 = first_graph.getEdges();
        undirectedGraphsGenerators::AllUndirectedGraphsGenerator generator2(n);

        do {
            index++;
            if (first_components == 1) {
                connected_index++;
                if (n >= 5 || (n == 4 && connected_index % 20 != 0)) {
                    continue;
                }
            }
            if (index < start) {
                continue;
            }

            auto second_graph = generator2.buildGraph();
            if (!validate_second_graph(second_graph)) {
                continue;
            }
            auto edges2 = second_graph.getEdges();
            if (!result.add_test_case(n, edges1, edges2)) {
                fail = true;
                break;
            }
        } while (generator2.next());
    } while (!fail && generator1.next());

    logger << index << '\n';
}

void gen_all_trees() {
    int n = parser.getInt("n");
    treesGenerators::AllTreesGenerator generator(n);

    do {
        if (!result.add_test_case(n, {}, generator.buildGraph().getEdges())) {
            global::logAndThrowException("Did not finish generation of all trees.");
        }
    } while (generator.next());

    for (bool fail = false; !fail;) {
        generator = treesGenerators::AllTreesGenerator(n);
        do {
            if (!result.add_test_case(n, generate_first_graph(n), generator.buildGraph().getEdges())) {
                fail = true;
            }
        } while(!fail);
    }
}

void gen_all_stars() {
    int n = parser.getInt("n");
    undirectedGraphsGenerators::AllUndirectedGraphsGenerator generator(n);

    vector<pair<int, int>> second_edges;
    for (int i = 1; i < n; i++) {
        second_edges.emplace_back(0, i);
    }

    do {
        auto first_graph = generator.buildGraph();
        if (first_graph.components() == 1) {
            continue;
        }
        if (!validate_first_graph(first_graph)) {
            continue;
        }
        if (validate_second_graph(UndirectedGraph(n, second_edges)) && !result.add_test_case(n, first_graph.getEdges(), second_edges)) {
            global::logAndThrowException("Did not generate all graphs");
        }

        bool no_bad_edges = true;
        for (int i = 1; i < n && no_bad_edges; i++) {
            no_bad_edges &= !first_graph.containsEdge({0, i});
        }
        if (!no_bad_edges) {
            continue;
        }
        vector<pair<int, int>> complete_second_graph;
        for (int i = 0; i < n; i++) {
            for (int j = i + 1; j < n; j++) {
                if (!first_graph.containsEdge({i, j})) {
                    complete_second_graph.emplace_back(i, j);
                }
            }
        }
        if (validate_second_graph(UndirectedGraph(n, complete_second_graph)) && !result.add_test_case(n, first_graph.getEdges(), complete_second_graph)) {
            global::logAndThrowException("Did not generate all graphs");
        }
    } while (generator.next());
}

void gen_all_bamboos() {
    int n = parser.getInt("n");
    vector<pair<int, int>> second_edges;
    for (int i = 1; i < n; i++) {
        second_edges.emplace_back(i - 1, i);
    }

    undirectedGraphsGenerators::AllUndirectedGraphsGenerator generator(n);
    do {
        auto first_graph = generator.buildGraph();
        if (first_graph.components() == 1 || !validate_first_graph(first_graph)) {
            continue;
        }
        if (!result.add_test_case(n, first_graph.getEdges(), second_edges)) {
            global::logAndThrowException("Did not generate all graphs");
        }
    } while (generator.next());
}

void gen_independent() {
    const int tests = parser.getInt("tests");
    const int sum_n = parser.getInt("n", result.MAX);
    for (auto n : random::distributeSize(sum_n, 3, tests, rnd)) {
        if (!result.add_test_case(n, generate_first_graph(n), generate_second_graph(n))) {
            logger << "Did not finish to generate all test cases.\n";
            break;
        }
    }
}

void gen_complete_xor() {
    const int n = parser.getInt("n");

    auto generate_another_graph = [&](const UndirectedGraph &graph) {
        vector<pair<int, int>> edges;
        edges.reserve(n * (n - 1) / 2 - graph.size().second);
        for (int i = 0; i < n; i++) {
            for (int j = i + 1; j < n; j++) {
                if (!graph.containsEdge({i, j})) {
                    edges.emplace_back(i, j);
                }
            }
        }
        return edges;
    };

    if (parser.hasParameter("first_mode")) {
        while (true) {
            auto first_graph = generate_first_graph(n);
            auto second_graph = generate_another_graph(UndirectedGraph(n, first_graph));
            if (!result.add_test_case(n, std::move(first_graph), std::move(second_graph))) {
                break;
            }
        }
    } else if (parser.hasParameter("second_mode")) {
        while (true) {
            auto second_graph = generate_second_graph(n);
            auto first_graph = generate_another_graph(UndirectedGraph(n, second_graph));
            if (!result.add_test_case(n, std::move(first_graph), std::move(second_graph))) {
                break;
            }
        }
    } else {
        global::logAndThrowException("No first_mode and second_mode are provided");
    }
}

void gen_no_common_edges() {
    const int sum_n = parser.getInt("n");
    const int tests = parser.getInt("tests");
    for (auto n : random::distributeSize(sum_n, 3, tests, rnd)) {
        auto second_graph = UndirectedGraph(n, generate_second_graph(n));
        int iterations = n * parser.getDouble("edges_frac");
        set<pair<int, int>> was;
        while (iterations--) {
            int v, u;
            do {
                u = random::random(0, n, rnd);
                v = random::random(0, n, rnd);
            } while (v == u);
            if (second_graph.containsEdge({v, u}) || !was.insert(minmax(v, u)).second) {
                continue;
            }
        }

        if (!result.add_test_case(n, vector<pair<int, int>>(was.begin(), was.end()), second_graph.getEdges())) {
            global::logAndThrowException("Did not finish the generation.");
        }
    }
}

int main(int argc, char *argv[]) {
    registerGen(argc, argv, 1);
    parser = ArgumentParser(argc, argv);

    result.group = parser.getInt("group");
    result.MAX = MAX_N_M[result.group];

    string gen_mode = parser.getString("gmode");
    if (gen_mode == "all_cases") {
        gen_all_possible_cases();
    } else if (gen_mode == "all_trees") {
        gen_all_trees();
    } else if (gen_mode == "all_stars") {
        gen_all_stars();
    } else if (gen_mode == "all_bamboos") {
        gen_all_bamboos();
    } else if (gen_mode == "independent") {
        gen_independent();
    } else if (gen_mode == "complete_xor") {
        gen_complete_xor();
    } else if (gen_mode == "no_common_edges") {
        gen_no_common_edges();
    } else {
        assert(false);
    }

    result.print();
}
