from dataclasses import dataclass
from typing      import Callable

import subprocess
import sys

class GroupsHandler:
    GROUP: int = 0

    @dataclass
    class Group:
        group:     int
        score:     int
        depends:   list[int]
        groupExec: Callable[None, None]

    def __init__(self):
        self.groups = []

    def registerGroup(self, group: int, score: int=0, depends: list[int]=None):
        if depends is None:
            depends = []

        def decoratorOuter(groupExec):
            self.groups.append(GroupsHandler.Group(group, score, depends, groupExec))
            return groupExec

        return decoratorOuter

    def execute(self):
        self.groups.sort(key=lambda x: x.group)
        for groupData in self.groups:
            GroupsHandler.printGroupHeader(groupData)
            GroupsHandler.GROUP = groupData.group
            groupData.groupExec()

# Private:
    @staticmethod
    def printGroupHeader(groupData: Group):
        print(f'<#-- group {groupData.group} score {groupData.score}', end=' ')
        if (len(groupData.depends) > 0):
            print('depends', end=' ')
            groupData.depends.sort()
            for index in groupData.depends:
                print(index, end=' ')
        print('-->')

def print_script(script):
    print(script, '> $')

def add_script(flags: list[str] = None, **kwargs):
    if flags is None:
        flags = []
    for key, value in kwargs.items():
        flags.append(f'{key}={value}')
    print_script(f'gen group={GroupsHandler.GROUP} ' + ' '.join(flags))

def execute(flags: list[str] = None, **kwargs):
    if flags is None:
        flags = []
    for key, value in kwargs.items():
        flags.append(f'{key}={value}')

    script = f'gen group={GroupsHandler.GROUP} ' + ' '.join(flags)
    result = subprocess.run(['./gen'] + flags, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    error = result.stderr.decode()
    if result.returncode != 0:
        print('Error occured inside generator.')
        print(f'script: {script}')
        print(error)
        sys.exit(1)

    return (error, script)

def generate_all_graphs(n, maxIterations=100):
    index = 0
    for i in range(maxIterations):
        (new_index, script) = execute(group=GroupsHandler.GROUP, gmode="all_cases", n=n, graph_start=index)
        new_index = int(new_index)
        if new_index == index:
            break

        index = new_index
        print_script(script)
