#include "testlib.h"
#include <bits/stdc++.h>
using namespace std;

constexpr int N = 3'000'00;
constexpr int MAX_TESTS = 60'000;
constexpr int SMALL_N = 3000;
constexpr int GROUPS = 10;
constexpr int MAX_N_M[GROUPS + 1]{/* 0:  */ N,
                                  /* 1:  */ 5,
                                  /* 2:  */ SMALL_N,
                                  /* 3:  */ SMALL_N,
                                  /* 4:  */ SMALL_N,
                                  /* 5:  */ SMALL_N,
                                  /* 6:  */ SMALL_N,
                                  /* 7:  */ SMALL_N,
                                  /* 8:  */ SMALL_N,
                                  /* 9:  */ N,
                                  /* 10: */ N};

struct disjoin_set_union {
    vector<int> par;

    disjoin_set_union(int n) : par(n) {
        iota(par.begin(), par.end(), 0);
    }

    int root(int v) {
        return par[v] == v ? v : (par[v] = root(par[v]));
    }

    bool unite(int v, int u) {
        if (root(v) == root(u)) return false;
        par[root(v)] = root(u);
        return true;
    }
};

int get_components(int n, const vector<pair<int, int>> &edges) {
    int components = n;
    disjoin_set_union dsu(n);
    for (const auto &[v, u] : edges) {
        components -= dsu.unite(v, u);
    }
    return components;
}

vector<int> get_degs(int n, const vector<pair<int, int>> &edges) {
    vector<int> degs(n);
    for (const auto &[v, u] : edges) {
        degs[v]++, degs[u]++;
    }
    return degs;
}

bool find_edge(pair<int, int> edge, const vector<pair<int, int>> &edges) {
    for (int rot = 0; rot < 2; rot++, swap(edge.first, edge.second)) {
        if (find(edges.begin(), edges.end(), edge) != edges.end()) {
            return true;
        }
    }
    return false;
}

void validate_first_graph(int n, const vector<pair<int, int>> &edges, int group) {
    int components = get_components(n, edges);
    if (group == 3) {
        ensuref(components == 2, "First graph contains %d components instead of 2 in group #%d.", components, group);
    }
    if (group == 4) {
        auto degs = get_degs(n, edges);
        ensuref(*min_element(degs.begin(), degs.end()) >= 1, "Degree of each vertex must be at least 1 in group #%d.", group);
    }
    if (group == 7) {
        ensuref(find_edge({0, 1}, edges), "First graph does not contain edge (1, 2) in group #%d", group);
    }
}

void validate_second_graph(int n, const vector<pair<int, int>> &edges, int group) {
    int components = get_components(n, edges);
    ensuref(components == 1, "Second graph contains %d components instead of 1.", components);

    if (group == 2) {
        ensuref(int(edges.size()) == 1ll * n * (n - 1) / 2, "Second graph must be complete in group #%d.", group);
    }
    if (group == 9 || group == 5 || group == 6) {
        ensuref(components == 1 && int(edges.size()) == n - 1, "Second graph must be a tree in group #%d.", group);
    }

    if (group == 9) {
        for (int i = 0; i + 1 < n; i++) {
            ensuref(edges[i] == pair<int, int>{i, i + 1}, "Second graph does not contain edge (%d, %d) in group #%d.", i + 1, i + 2, group);
        }
    }
    if (group == 5) {
        for (int i = 0; i + 1 < n; i++) {
            ensuref(edges[i] == pair<int, int>{0, i + 1}, "Second graph does not contain edge (%d, %d) in group #%d.", 1, i + 2, group);
        }
    }

    if (group == 7) {
        ensuref(find_edge({0, 1}, edges), "Second graph does not contain edge (1, 2) in group #%d", group);
    }
}

int main(int argc, char* argv[]) {
    registerValidation(argc, argv);
    int group = atoi(validator.group().c_str());

    int tests = inf.readInt(1, MAX_TESTS, "t");
    inf.readSpace();
    inf.readInt(group, group, "g");
    inf.readEoln();

    const int MAX_N = MAX_N_M[group];
    const int MAX = (group == 1 ? SMALL_N : MAX_N);
    int sum_n = 0;
    int sum_m1 = 0;
    int sum_m2 = 0;
    for (int test_index = 1; test_index <= tests; test_index++) {
        setTestCase(test_index);

        int n = inf.readInt(3, MAX_N, "n");
        inf.readSpace();
        int m1 = inf.readInt(0, MAX, "m1");
        inf.readSpace();
        int m2 = inf.readInt(0, MAX, "m2");
        inf.readEoln();

        sum_n += n;
        sum_m1 += m1;
        sum_m2 += m2;
        ensuref(sum_n <= MAX, "Sum of n exeeds %d in group #%d.", MAX, group);
        ensuref(max(sum_m1, sum_m2) <= MAX, "Sum of m exeeds %d in group #%d.", MAX, group);

        auto read_graph = [&](int m) {
            vector<pair<int, int>> edges(m);
            set<pair<int, int>> st;
            for (auto &[v, u] : edges) {
                v = inf.readInt(1, n, "v") - 1;
                inf.readSpace();
                u = inf.readInt(1, n, "u") - 1;
                inf.readEoln();
                ensuref(v != u, "Self-loop in the graph.");
                ensuref(st.insert(minmax(v, u)).second, "Multiple edge in the graph.");
            }
            return edges;
        };

        validate_first_graph(n, read_graph(m1), group);
        validate_second_graph(n, read_graph(m2), group);
    }
    inf.readEof();
}
