#include <bits/stdc++.h>
using namespace std;

struct DSU {
    vector<int> par;

    DSU(int n) : par(n) {
        iota(par.begin(), par.end(), 0);
    }
    
    int root(int v) {
        return par[v] == v ? v : (par[v] = root(par[v]));
    }

    void unite(int v, int u) {
        par[root(v)] = root(u);
    }
};

void solve() {
    int n, m1, m2;
    cin >> n >> m1 >> m2;

    DSU dsu(n);
    set<pair<int, int>> edges1;
    for (int i = 0; i < m1; i++) {
        int v, u;
        cin >> v >> u;
        edges1.insert(minmax(v - 1, u - 1));
        dsu.unite(v - 1, u - 1);
    }
    
    set<pair<int, int>> edges2;
    for (int i = 0; i < m2; i++) {
        int v, u;
        cin >> v >> u;
        edges2.insert(minmax(v - 1, u - 1));
    }
    vector<pair<int, int>> ops;

    auto output = [&]() {
        cout << "Yes\n";
        cout << ops.size() << '\n';
        for (auto [v, u] : ops) {
            cout << v + 1 << ' ' << u + 1 << '\n';
        }
    };

    auto solve_pair = [&](int v, int u) { 
        assert(dsu.root(v) == dsu.root(u));
        for (int i = 0; i < n; i++) {
            if (dsu.root(i) == i && dsu.root(i) != dsu.root(v)) {
                ops.emplace_back(i, edges2.count(minmax(v, i)) ? u : v);
            }
        }
        output();
    };

    for (auto [v, u] : edges2) {
        if (dsu.root(v) == dsu.root(u)) {
            solve_pair(v, u);
            return;
        }
    }

    for (int i = 0; i < n; i++) {
        for (int j = i + 1; j < n; j++) {
            if (edges1.count({i, j}) || edges2.count({i, j})) continue;
            ops.emplace_back(i, j);
            dsu.unite(i, j);
            solve_pair(i, j);
            return;
        }
    }

    /*
     * Every component in the first graph is a clique.
     * There are all edges between all components of the first graph in the second graph.
     */

    if (n == 3) {
        cout << "No\n"; // lol ;)
        return;
    }

    if (m1 == 0) {
        dsu.unite(0, 1);
        ops.emplace_back(0, 1);
    }

    vector<vector<int>> components(n);
    for (int i = 0; i < n; i++) {
        components[dsu.root(i)].push_back(i);
    }
    sort(components.begin(), components.end(), [](const auto &a, const auto &b) {
        return a.size() > b.size();
    });
    assert(components[0].size() >= 2);

    if (int(components[0].size()) == n - 1) {
        ops.emplace_back(components[0][0], components[0][1]);
    }
    for (int i = 1; i < n; i++) {
        if (!components[i].empty()) {
            ops.emplace_back(components[i][0], components[0][i == 1]);
        }
    }
    output();
}

int main() {
    cin.tie(nullptr)->sync_with_stdio(false);
    int tests, group;
    cin >> tests >> group;
    while (tests--) {
        solve();
    }
}
